#!/usr/bin/python3
import csv
import math
import cmath
import time
import numpy
from argparse import ArgumentParser
import pandas as pd
import matplotlib.pyplot as plt
import climate_chamber
import VNA
import virtual_time
import json


TEMPERATURE_STABLE = 0x1
HUMIDITY_STABLE = 0x2
MAGNITUDE_STABLE = 0x4
PHASE_STABLE = 0x8
MEASUREMENT_STABLE = 0x10


# Class has only attributes which we are using in the read_data_function to read data from VNA and Chamber
class MeasurementData:
    def __init__(self, timestamp, temp, hum, power, frequency, s11, s21, s12, s22):
        self.timestamp = timestamp
        self.temp = temp
        self.hum = hum
        self.power = power
        self.frequency = frequency
        self.s11 = s11
        self.s21 = s21
        self.s12 = s12
        self.s22 = s22


class Measurements:
    def __init__(self, chamber_address, vna_address, sweep_file, output_file, standby, config_data):
        self.max_delta_temp = config_data['delta_temp']
        self.max_delta_hum = config_data['delta_hum']
        self.max_delta_mag = config_data['delta_mag']
        self.max_delta_phase = config_data['delta_phase']
        self.sleep_time = config_data["sleep_time"]
        self.frequency = config_data["frequency"]
        target_accuracy = [self.max_delta_temp, self.max_delta_hum]
        self.chamber = climate_chamber.create_chamber(chamber_address, target_accuracy)
        self.vna = VNA.create_vna(vna_address, target_accuracy)
        self.sweep_file = sweep_file
        self.standby = standby
        self.output = output_file
        self.clock = virtual_time.get_clock(chamber_address, target_accuracy)
        self.temperature_stable = False
        self.humidity_stable = False
        self.magnitude_stable = False
        self.phase_stable = False

        self.vna.create_new_trace("Trace1", "S11")
        self.vna.create_new_trace("Trace2", "S11")
        self.vna.create_new_trace("Trace3", "S21")
        self.vna.create_new_trace("Trace4", "S21")
        self.vna.create_new_trace("Trace5", "A1")
        self.vna.create_new_trace("Trace6", "B1")
        self.vna.create_new_trace("Trace7", "A2")
        self.vna.create_new_trace("Trace8", "B2")
        self.vna.set_cw_frequency(self.frequency)
        self.vna.set_sweep_mode('CW')

    def perform_measurements(self):
        with open(self.output, mode='w', newline='') as csv_file:
            fieldnames = ['TIMESTAMP', 'TARGET_TEMPERATURE', 'READBACK_TEMPERATURE', 'TARGET_HUMIDITY',
                          'READBACK_HUMIDITY', 'RF_POWER', 'RF_FREQUENCY', 'DUT_IDENTIFIER', 'RUN_ID',
                          'EQUILIBRIUM_INDICATOR', 'S11_MAGNITUDE', 'S11_PHASE', 'S12_MAGNITUDE', 'S12_PHASE',
                          'S21_MAGNITUDE', 'S21_PHASE', 'S22_MAGNITUDE', 'S22_PHASE']
            # csv.dict writer add adda row wise
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()

            with open(self.sweep_file) as file:
                try:
                    while line := file.readline().rstrip():
                        list_of_values = line.split()
                        next_temp = float(list_of_values[0])  # target_temperature
                        next_hum = float(list_of_values[1])  # target_humidity
                        next_soaking = int(list_of_values[2])  # soaking_time
                        next_reads = int(list_of_values[3])  # number of stable readings

                        set_const_response = self.chamber.set_const((next_temp, next_hum))
                        print(set_const_response)
                        set_mode_response = self.chamber.set_mode('CONSTANT')
                        print(set_mode_response)

                        number_of_soaking_reads = next_soaking / self.sleep_time + 1

                        do_another_measurement = True
                        while do_another_measurement:
                            # wait until set point is reached (+soaking time)
                            magnitudes_queue = []
                            phase_queue = []

                            while True:
                                data = self.read_data()
                                # if it is not within the target range reset start time
                                self.temperature_stable = False
                                self.humidity_stable = False

                                self.temperature_stable = self.calculate_temperature_stability(next_temp, float(data.temp))
                                self.humidity_stable = self.calculate_humidity_stability(next_hum, float(data.hum))

                                # The queue must not be longer than the max number of soaking reads.
                                # If the queue is already full, we have to pop the first element before we can add the
                                # current measurement.
                                if len(magnitudes_queue) >= number_of_soaking_reads:
                                    magnitudes_queue.pop(0)
                                if self.temperature_stable and self.humidity_stable:
                                    magnitudes_queue.append(self.calculate_mean_magnitude(data.s21))
                                else:
                                    magnitudes_queue.clear()
                                # check cable stability parameters
                                self.magnitude_stable = False
                                if len(magnitudes_queue) >= number_of_soaking_reads:
                                    spread = max(magnitudes_queue) - min(magnitudes_queue)
                                    if spread < 2*self.max_delta_mag:
                                        self.magnitude_stable = True

                                if len(phase_queue) >= number_of_soaking_reads:
                                    phase_queue.pop(0)
                                if self.temperature_stable and self.humidity_stable:
                                    phase_queue.append(self.calculate_mean_phase(data.s21))
                                else:
                                    phase_queue.clear()

                                self.phase_stable = False
                                if len(phase_queue) >= number_of_soaking_reads:
                                    spread = max(phase_queue) - min(phase_queue)
                                    if spread < 2*self.max_delta_phase:
                                        self.phase_stable = True

                                print('Setpoint: ' + str(next_temp) + ' ' + str(next_hum) + ' | Temp: ' + data.temp +
                                      ' Â°C' + ' | Humid: ' + data.hum + '%'
                                      + ' | soaking read nr' + str(len(magnitudes_queue)))
                                self.write_data(writer, next_temp, next_hum, data, self.cook_up_equi_indicator())

                                if self.temperature_stable and self.humidity_stable and self.magnitude_stable and\
                                        self.phase_stable:
                                    reference_magnitude = magnitudes_queue[-1]
                                    reference_phase = phase_queue[-1]
                                    print('SOAKING FINISHED!')
                                    break
                                else:
                                    self.clock.sleep(self.sleep_time)

                            # perform the configured number of measurements and check that they are really stable
                            # It started running after everything become stable
                            supposedly_stable_measurements = []
                            all_measurements_stable = True
                            for i in range(0, next_reads):
                                data = self.read_data()
                                self.temperature_stable = self.calculate_temperature_stability(next_temp, float(data.temp))
                                self.humidity_stable = self.calculate_humidity_stability(next_hum, float(data.hum))
                                mag = self.calculate_mean_magnitude(data.s21)
                                phase = self.calculate_mean_phase(data.s21)
                                self.magnitude_stable = (reference_magnitude-self.max_delta_mag <= mag) and\
                                                        (mag <= reference_magnitude+self.max_delta_mag)
                                self.phase_stable = (reference_phase-self.max_delta_phase <= phase) and\
                                                    (phase <= reference_phase+self.max_delta_phase)

                                supposedly_stable_measurements.append([data, self.cook_up_equi_indicator()])
                                
                                if (self.temperature_stable and self.humidity_stable and self.magnitude_stable and
                                        self.phase_stable):
                                    print('Stable measurement ' + str(i+1) + '/' + str(next_reads))
                                    self.clock.sleep(self.sleep_time)
                                else:
                                    all_measurements_stable = False
                                    print('Measurement not stable. Retrying.')
                                    break

                            for [d, measurement_equi] in supposedly_stable_measurements:
                                if all_measurements_stable:
                                    equi = TEMPERATURE_STABLE | HUMIDITY_STABLE | MAGNITUDE_STABLE | PHASE_STABLE | MEASUREMENT_STABLE
                                    do_another_measurement = False
                                else:
                                    equi = measurement_equi
                                self.write_data(writer, next_temp, next_hum, d, equi)

                except KeyboardInterrupt:
                    pass

        if self.standby:
            standby_response = self.chamber.set_mode('STANDBY')
            print(standby_response)

    def read_data(self):
        [temp, hum, mode, alarms] = self.chamber.read_monitor().split(',')
        power = self.vna.get_current_power()
        frequency = self.vna.get_current_cw_frequency()
        s11 = self.get_trace_data("Trace1")
        s21 = self.get_trace_data("Trace2")
        s12 = self.get_trace_data("Trace3")
        s22 = self.get_trace_data("Trace4")

        return MeasurementData(int(self.clock.time()), temp, hum, power, frequency, s11, s21, s12, s22)

    def write_data(self, writer, target_temp, target_hum, data, equi_indicator):
        writer.writerow({
            'TIMESTAMP': data.timestamp,
            'TARGET_TEMPERATURE': target_temp,
            'READBACK_TEMPERATURE': data.temp,
            'TARGET_HUMIDITY': target_hum,
            'READBACK_HUMIDITY': data.hum,
            'RF_POWER': data.power,
            'RF_FREQUENCY': data.frequency,
            'EQUILIBRIUM_INDICATOR': equi_indicator,
            'S11_PHASE': self.calculate_mean_phase(data.s11),
            'S11_MAGNITUDE': self.calculate_mean_magnitude(data.s11),
            'S21_PHASE': self.calculate_mean_phase(data.s21),
            'S21_MAGNITUDE': self.calculate_mean_magnitude(data.s21),
            'S12_PHASE': self.calculate_mean_phase(data.s12),
            'S12_MAGNITUDE': self.calculate_mean_magnitude(data.s12),
            'S22_PHASE': self.calculate_mean_phase(data.s22),
            'S22_MAGNITUDE': self.calculate_mean_magnitude(data.s22)
        })

    def current_milli_time(self):
        return int(round(self.clock.time() * 1000))

    def get_trace_data(self, trace):
        return self.vna.get_list_of_measurement_values(trace, "SDAT")

    def calculate_complex_numbers(self, values_list):
        real_num = values_list[::2]
        imaginary_num = values_list[1::2]
        complex_numbers = []
        for i, q in zip(real_num, imaginary_num):
            complex_numbers.append(complex(i, q))
        return complex_numbers

    def calculate_magnitudes(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        magnitudes = [abs(val) for val in complex_numbers]
        return magnitudes

    def calculate_mean_magnitude(self, values_list):
        magnitudes = self.calculate_magnitudes(values_list)
        return numpy.mean(magnitudes)

    def calculate_phases(self, values_list):
        complex_numbers = self.calculate_complex_numbers(values_list)
        phases = [math.degrees(cmath.phase(val)) for val in complex_numbers]
        return phases

    def calculate_mean_phase(self, values_list):
        phases = self.calculate_phases(values_list)
        return numpy.mean(phases)

    def cook_up_equi_indicator(self):
        equilibrium_indicator = 0

        if self.temperature_stable:
            equilibrium_indicator = equilibrium_indicator | TEMPERATURE_STABLE

        if self.humidity_stable:
            equilibrium_indicator = equilibrium_indicator | HUMIDITY_STABLE

        if self.magnitude_stable:
            equilibrium_indicator = equilibrium_indicator | MAGNITUDE_STABLE

        if self.phase_stable:
            equilibrium_indicator = equilibrium_indicator | PHASE_STABLE

        return equilibrium_indicator

    def plot_output(self, output_file, show_plot_window):
        csv_data = pd.read_csv(output_file)
        fig, ax1 = plt.subplots(2, figsize=(12, 10))
        fig.suptitle("Graphical representation of chamber output", color="red")

        path_collection01 = ax1[0].scatter(csv_data.TIMESTAMP, csv_data.S11_PHASE, c='red', marker='<', label='Phase')
        twin2_0 = ax1[0].twinx()
        path_collection02 = twin2_0.scatter(csv_data.TIMESTAMP, csv_data.S11_MAGNITUDE, c='#3120E0', marker='4',
                                            label='Magnitude')
        twin3_0 = ax1[0].twinx()
        twin3_0.spines['right'].set_position(('outward', 40))
        path_collection03 = twin3_0.scatter(csv_data.TIMESTAMP, csv_data.EQUILIBRIUM_INDICATOR, c='black', marker=".",
                                            label='Equilibrium_Indicator')
        ax1[0].set_xlabel("TIMESTAMP")
        ax1[0].set_ylabel("PHASE", color='red')
        twin2_0.set_ylabel("MAGNITUDE", color='#3120E0')
        twin3_0.set_ylabel("EQUILIBRIUM_INDICATOR", color='black')

        ax1[0].grid(True, linestyle=":")
        all_path_collections = [path_collection01, path_collection02, path_collection03]
        labels = [pc.get_label() for pc in all_path_collections]
        ax1[0].legend(all_path_collections, labels, loc='lower right')

        path_collection11 = ax1[1].scatter(csv_data.TIMESTAMP, csv_data.READBACK_TEMPERATURE, c='blue', marker='p',
                                           label="Temperature")
        twin2_1 = ax1[1].twinx()
        path_collection12 = twin2_1.scatter(csv_data.TIMESTAMP, csv_data.READBACK_HUMIDITY, c='green', marker="*",
                                            label="Humidity")
        twin3_1 = ax1[1].twinx()
        twin3_1.spines['right'].set_position(('outward', 40))
        path_collection13 = twin3_1.scatter(csv_data.TIMESTAMP, csv_data.EQUILIBRIUM_INDICATOR, c='black', marker=".",
                                            label="Equilibrium_Indicator")
        ax1[1].set_xlabel("TIMESTAMP")
        ax1[1].set_ylabel("TEMPERATURE ", color='blue')
        twin2_1.set_ylabel("HUMIDITY", color='green')
        twin3_1.set_ylabel("EQUILIBRIUM_INDICATOR", color='black')
        ax1[1].grid(True, linestyle=":")
        all_path_collections = [path_collection11, path_collection12, path_collection13]
        labels = [pc.get_label() for pc in all_path_collections]
        ax1[1].legend(all_path_collections, labels, loc='lower right')

        fig.savefig(time.strftime("%Y_%m_%d-%H_%M_%S") + 'graph.pdf')

        if show_plot_window:
            plt.show()

    def calculate_temperature_stability(self, target_temp, readback_temp):
        return (target_temp-self.max_delta_temp <= float(readback_temp)) and \
               (float(readback_temp) <= target_temp+self.max_delta_temp)

    def calculate_humidity_stability(self, target_hum, readback_hum):
        return (target_hum-self.max_delta_hum <= float(readback_hum)) and \
               (float(readback_hum) <= target_hum+self.max_delta_hum)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument("-c", "--chamber",
                        help="IP address of climate chamber", metavar="ADDR",
                        required=True)
    parser.add_argument("-v", "--vna",
                        help="IP address of VNA", metavar="ADDR",
                        required=True)
    parser.add_argument('-f', '--file',
                        help='File containing list of sweeps',
                        required=True)
    parser.add_argument('-o', '--output',
                        help='Name of output file')
    parser.add_argument("-s", "--standby",
                        action="store_true", default=False,
                        help="Change Chamber mode to standby at the end")
    parser.add_argument("-p", "--plot",
                        action="store_true", default=False,
                        help="Show plot window at the end")

    # parser.add_argument("-q", "--quiet",
    #                     action="store_false", dest="verbose", default=True,
    #                     help="don't print status messages to stdout")
    args = parser.parse_args()

    if not args.output:
        output = time.strftime("%Y_%m_%d-%H_%M_%S") + "_results.csv"
    else:
        output = args.output

    print(args.chamber, args.vna, args.file, output, args.standby)

    # reading json file for target accuracy
    with open('test_stand_parameter.json', 'r') as f:
        config_data = json.load(f)

    mes = Measurements(args.chamber, args.vna, args.file, output, args.standby, config_data)
    try:

        mes.perform_measurements()
        mes.plot_output(output, args.plot)

    finally:
        mes.chamber.close()

    # mes.plot_output(output)

